@@ -116,6 +116,32 @@ class Rigol_DSG800:
 #================================================================
 # Function Generator - Rigol DG4000 series
 #================================================================
+def power_vrms(v,z):
+    """
+    Calculate power using Vrms
+    Input: Vrms, impedance Z
+    Output: power (W)
+    """
+    return v**2/z
+
+def power_vpp(v,z):
+    """
+    Calculate power using Vpp
+    Input: Vpp, impedance Z
+    Output: power (W)
+    
+    """
+    return v**2/z/8
+
+def power_dbm(dbm):
+    """
+    Convert power from dBm to W
+    Input: dBm
+    Output: power (W)
+    
+    """
+    return pow(10, -3 + dbm/10)
+
 class Rigol_DG4000:
     def __init__(self,address):
         self.address = address
@@ -140,9 +166,9 @@ class Rigol_DG4000:
             self.dev.write(':OUTput1 ON') 
             self.dev.write(':OUTput2 ON')
 
-    def __Hz(f):  # in Hz, support unit. Default: MHz
+    def __Hz(self, f):  # in Hz, support unit. Default: MHz
         """
-        Description: Sets all frequencies to MHz Unit (Tested 04/03/2022)
+        Description: Sets all frequencies to MHz Unit (Tested 07/15/2022)
 
         Input: f : frequency : float
 
@@ -157,50 +183,146 @@ class Rigol_DG4000:
                 return 1*float(re.sub('[a-zA-Z]','',f))
         else: # float, or str that contains only numbers
             return float(f)*1e6
+        
+    def output_impedence(self, ch, *load):
+        if bool(load) == False:
+            load = 'INF'
+        else:
+            load = load[0]
+        self.dev.write(':OUTPut{}:LOAD {}'.format(ch,load))
 
     def freq(self,ch,*f):
         if bool(f) == True:   # assign value (only the first in the tuple)
-            __freq = f[0]
-            self.dev.write(':SOURCe{}:Freq {}'.format(ch,self.__Hz(__freq)));
+            freq = self.__Hz(f[0])
+            try:
+                freq_apply = min(freq,  eval('self.hz_max_{}'.format(ch)))
+            except:
+                freq_apply = freq
+            self.dev.write(':SOURCe{}:Freq {}'.format(ch, freq_apply));
+            if freq_apply < freq:
+                print('Warning: channel {} output frequency is too high. Reduced to {} Hz'.
+                      format(ch,eval('self.hz_max_{}'.format(ch))))
         else:   # when no input is entered, read actual frequency 
             __readOut = self.dev.query(':SOURCe{}:Freq?'.format(ch))
             __freq = float(re.search('.*(?=\n)',__readOut).group())
             return float(__freq)*1e-6  # MHz
         
+    def freq_max(self, ch, *f):
+        """
+        Setup the frequency limit.
+        
+        Input: channel, frequency max (any unit)
+        
+        Output: maximum frequency (Hz)
+        """
+        if bool(f):
+            hz_max = f[0]
+            exec("self.hz_max_{} = {} ".format(ch, self.__Hz(hz_max)))
+        else:
+            try:
+                return eval("self.hz_max_{}".format(ch))   # 'Hz'
+            except:
+                print('Warning: maximum output frequency has not been specified yet.' )
+                    
     def lev(self,ch,*v):
         if bool(v) == True:
             __v = v[0]
             if type(__v) == str:
                 __lev = float(re.sub('[a-zA-Z]','',__v))  # unitless value
-                if re.search('(vrms)',__v,re.IGNORECASE) != None:  #  VRMS
-                    self.dev.write(':SOURCe{}:VOLTage:UNIT VRMS'.format(ch))
-                elif re.search('d',__v,re.IGNORECASE) != None:  # dBm
-                    self.dev.write(':SOURCe{}:VOLTage:UNIT DBM'.format(ch))
-                else:  # VPP
-                    self.dev.write(':SOURCe{}:VOLTage:UNIT VPP'.format(ch))
                 # mVPP or mVRMS
                 if re.search('(mv)',__v,re.IGNORECASE) != None:
                     __lev = 1e-3 * __lev
-                # value
-                self.dev.write(':SOURCe{}:VOLTage {}'.format(ch,__lev))
+                # apply voltage 
+                if re.search('(vrms)',__v,re.IGNORECASE) != None:  #  VRMS
+                    self.dev.write(':SOURCe{}:VOLTage:UNIT VRMS'.format(ch))
+                    try:
+                        lev_apply = min(__lev, eval('self.vpp_max_{}/2/pow(2,1/2)'.format(ch)) )
+                    except:
+                        lev_apply = __lev
+                elif re.search('(dbm)',__v,re.IGNORECASE) != None:  # dBm
+                    self.dev.write(':SOURCe{}:VOLTage:UNIT DBM'.format(ch))
+                    try:
+                        lev_apply = min(__lev, eval('10*np.log10(1000*self.power_max_{})'.format(ch)) )
+                    except:
+                        lev_apply = __lev
+                elif re.search('(vpp)',__v,re.IGNORECASE) != None:  # VPP
+                    self.dev.write(':SOURCe{}:VOLTage:UNIT VPP'.format(ch))
+                    try:
+                        lev_apply = min(__lev, eval('self.vpp_max_{}'.format(ch)) )
+                    except:
+                        lev_apply = __lev
+                self.dev.write(':SOURCe{}:VOLTage {}'.format(ch,lev_apply))
             else:  # default: [Vpp] 
+                __lev = __v
+                try:
+                    lev_apply = min(__lev, eval('self.vpp_max_{}'.format(ch)) )
+                except:
+                    lev_apply = __lev
                 self.dev.write(':SOURCe{}:VOLTage:UNIT VPP'.format(ch))
-                self.dev.write(':SOURCe{}:VOLTage {}'.format(ch,__v))
+                self.dev.write(':SOURCe{}:VOLTage {}'.format(ch,lev_apply))
+            # print out warning
+            if lev_apply < __lev:
+                print('Protection Warning: output {} amplitude is too high. Reduced to limit value.'.format(ch))
         else:
             __readOut = self.dev.query(':SOURCe{}:VOLTage?'.format(ch))
             __lev = float(re.search('.*(?=\n)',__readOut).group())
             __readOut = self.dev.query(':SOURCe{}:VOLTage:UNIT?'.format(ch))
             __unit = re.search('.*(?=\n)',__readOut).group()
             return __lev, __unit
-
+        
+    def lev_max(self,ch, *v):
+        """
+        Set maximum output level, in any units.
+        * Note that Vrms - Vpp conversion only holds for sinusoidal wave!
+        Input: channel, maximum voltage/power.
+        
+        Output: maximum amplitude, (and maximum power if imepance is not inf).
+        """
+        z_string = self.dev.query(':OUTPut{}:LOAD?'.format(ch))
+        z = float(re.search('.*(?=\n)',z_string).group())
+        # assign max voltage
+        if bool(v) == True:  
+            __v = v[0]
+            if type(__v) == str:  # with unit
+                lev = float(re.sub('[a-zA-Z]','',__v))  # unitless value
+                if re.search('(vrms)',__v,re.IGNORECASE) != None:  #  VRMS
+                    exec("self.vpp_max_{} = lev * 2 * pow(2, 1/2) ".format(ch))
+                    if z != np.inf:
+                        exec("self.power_max_{} = power_vrms(lev,z)".format(ch))
+                elif re.search('(vpp)',__v,re.IGNORECASE) != None:  # Vpp
+                    exec("self.vpp_max_{} = lev".format(ch))
+                    if z != np.inf:
+                        exec("self.power_max_{} = power_vpp(lev,z)".format(ch))
+                elif (re.search('(dbm)',__v,re.IGNORECASE) != None):  # dbm
+                    exec("self.power_max_{} = power_dbm(lev)".format(ch))   # W
+                    if  (z != np.inf):
+                        exec("self.vpp_max_{} = 2*pow(2,1/2)*pow(self.power_max_{} * z, 1/2)"
+                             .format(ch,ch))
+                # mVPP or mVRMS
+                if re.search('(mv)',__v,re.IGNORECASE) != None:
+                    exec("self.vpp_max_{} = 1e-3 * self.vpp_max_{}".format(ch,ch))
+                    exec("self.power_max_{} = 1e-6 * self.power_max_{}".format(ch,ch))
+            else:  # default: [Vpp] 
+                exec("self.vpp_max_{} = v[0]".format(ch))
+                exec("self.power_max_{} = power_vpp(v[0],z)".format(ch))
+        else:  # read self.power_max
+            try:
+                return eval("self.vpp_max_{}".format(ch))   # 'Vpp'
+            except:
+                try:
+                    if bool(eval("self.power_max_{}".format(ch))) == True:
+                        print('Warning: maximum output voltage cannot be specified because of infinite impedance.' )
+                except:
+                    print('Warning: maximum output level has not been specified yet.' )
+        
     def offset(self,ch,offset):  # V_DC
-        self.dev.write(':SOURCe{}:VOLTage:OFFSet {}'.format(ch,offset));
+        self.dev.write(':SOURCe{}:VOLTage:OFFSet {}'.format(ch,offset))
 
-    def phase(self,ch,p):
-        self.dev.write(':SOURCe{}:PHASe {}'.format(ch,p));
+    def phase(self,ch,phase):
+        self.dev.write(':SOURCe{}:PHASe {}'.format(ch,phase))
         
     @staticmethod    
-    def gaussian(t,mu,FWHM,a): #Gaussian Function. Inputs: (FWHM, Amplitude, Center)
+    def gaussian(t,mu,FWHM,a): 
         sigma = (FWHM)/(2*np.sqrt(2*np.log(2)))
         amplitude = np.sqrt(2*np.pi*sigma**2)*a
         return amplitude*( 1/(sigma * np.sqrt(2*np.pi) ) )*np.exp( -((t-mu)**2 / (2*sigma**2)) )
@@ -224,7 +346,7 @@ class Rigol_DG4000:
         
     def arbitrary(self, ch, signal_width, waveform, *arg):
         """
-        Description: Allows one to set and create arbitrary waveform output (Tested 04/03/2022) 
+        Description: Allows one to set and create arbitrary waveform output (Tested 05/27/2022) 
 
         Input: ch : channel : int
                signal_width : width (time) of the argument in seconds : float
@@ -236,18 +358,22 @@ class Rigol_DG4000:
         buffer_size = 2**14
         if inspect.ismethod(waveform) == True or inspect.isfunction(waveform) == True:
             t = np.linspace(0,signal_width,buffer_size)
-            data = np.around(waveform(t,*arg),4)
-            datastring = ",".join(map(str,self.normalize(data)))
+            data = waveform(t,*arg)
+            midpoint = (max(data)-min(data))/2
+            centered_data = data - midpoint
+            datastring = ",".join(map(str,np.round(self.normalize(centered_data),4)))
         else:
-            data = np.around(waveform,4)
-            datastring = ",".join(map(str,self.normalize(data)))
-        
-        factor = max([np.abs(max(data)),np.abs(min(data))])
+            data = waveform
+            midpoint = (max(data)-min(data))/2
+            centered_data = data - midpoint
+            datastring = ",".join(map(str,np.round(self.normalize(centered_data),4)))
+    
         self.dev.write('SOURCE{}:Freq {}'.format(ch, 1/signal_width))
         self.dev.write("SOURCE{}:TRACE:DATA VOLATILE,".format(ch) + datastring)
         self.dev.write("SOURCE{}:FUNC VOLATILE,".format(ch))
-        self.dev.write("SOURCE{}:VOLTAGE:LOW {}".format(ch,-factor))
-        self.dev.write("SOURCE{}:VOLTAGE:HIGH {}".format(ch,factor))
+        self.dev.write("SOURCE{}:VOLTAGE:LOW {}".format(ch,min(centered_data)))
+        self.dev.write("SOURCE{}:VOLTAGE:HIGH {}".format(ch,max(centered_data)))
+        self.dev.write('SOURCe{}:VOLTage:OFFSet {}'.format(ch,midpoint))
         self.dev.write("SOURCE{}:PHASE:SYNC".format(ch))
     
     def burst(self, ch, mode, cycles):
@@ -317,7 +443,6 @@ class Rigol_DG4000:
         self.dev.write("SOURCE{}:BURST:NCYC {}".format(ch,cycles))
         self.dev.write("SOURCE{}:BURST:MODE:TRIG".format(ch))
         self.ext_trig(ch)
-
         
         
 #===========================================================================
@@ -818,8 +943,8 @@ class tektronix_AFG3000:
     def on(self): # default: both 
         self.dev.write(':OUTPut ON')
 
-    def freq(self,f):
-        self.dev.write('SOURce:FREQuency:FIXed {}'.format(self.__Hz(f)));
+    def freq(self,ch,f):
+        self.dev.write('SOURce{}:FREQuency:FIXed {}'.format(ch,self.__Hz(f)));
 
     def __Hz(self, f): 
         """
@@ -839,28 +964,28 @@ class tektronix_AFG3000:
         else: # float, or str that contains only numbers
             return float(f)*1e6
 
-    def lev(self,v):
+    def lev(self,ch,v):
         if type(v) == str:
             __V = float(re.sub('[a-zA-Z]','',v))  # unitless value
             if re.search('[rR]',v) != None:  #  VRMS
-                self.dev.write('SOURCe:VOLTage:UNIT VRMS')
+                self.dev.write('SOURCe:VOLTage:UNIT VRMS'.format(ch))
             elif re.search('[dD]',v) != None:  # dBm
-                self.dev.write('SOURCe:VOLTage:UNIT DBM')
+                self.dev.write('SOURCe:VOLTage:UNIT DBM'.format(ch))
         else:  # default: [Vpp] 
             __V = v
-            self.dev.write(':SOURCe:VOLTage:UNIT VPP')
+            self.dev.write(':SOURCe{}:VOLTage:UNIT VPP'.format(ch))
 
-        self.dev.write('SOURCe:VOLTage {}'.format(__V))
-        self.dev.write('SOURCe:VOLTage:UNIT VPP')
+        self.dev.write('SOURCe{}:VOLTage {}'.format(ch,__V))
+        self.dev.write('SOURCe{}:VOLTage:UNIT VPP')
 
-    def offset(self,offset):  # V_DC
-        self.dev.write('SOURce:VOLTage:LEVel:IMMediate:OFFSet {}'.format(offset));
+    def offset(self,ch,offset):  # V_DC
+        self.dev.write('SOURce{}:VOLTage:LEVel:IMMediate:OFFSet {}'.format(ch,offset));
 
-    def phase(self,phase):
-        self.dev.write('SOURC:PHASe {}'.format(phase));
+    def phase(self,ch,phase):
+        self.dev.write('SOURce{}:PHASe {}'.format(ch,phase));
         
-    def burst_delay(self,tdelay):
-        self.dev.write('SOUR:BURS:TDEL {}ns'.format(tdelay))
+    def burst_delay(self,ch,tdelay):
+        self.dev.write('SOURce{}:BURS:TDEL {}ns'.format(ch,tdelay))
     
     @staticmethod
     def gaussian(t,mu,FWHM,a): #Gaussian Function. Inputs: (Center, FWHM, Amplitude)
@@ -885,7 +1010,7 @@ class tektronix_AFG3000:
         factor = max([np.abs(max(waveform)),np.abs(min(waveform))])
         return np.array(waveform)/np.absolute(factor)
     
-    def arbitrary(self, signal_width, waveform, *arg):
+    def arbitrary(self, ch, signal_width, waveform, *arg):
         """
         Description: Allows one to set and create arbitrary waveform output 
 
@@ -926,15 +1051,15 @@ class tektronix_AFG3000:
             
         self.dev.write('DATA:DEFine EMEMory,{}'.format(len(data)))
         self.dev.write_binary_values("DATA:DATA EMEM1,", dac_values, datatype="H", is_big_endian=True)
-        self.dev.write("SOURce:FUNC:SHAPE EMEM1")
-        self.dev.write("SOURce1:VOLTage:LEVel:IMMediate:LOW {}".format(min(data)))
-        self.dev.write("SOURce1:VOLTage:LEVel:IMMediate:HIGH {}".format(max(data)))
-        self.dev.write("SOURCE:FREQ {}".format(self.__Hz(1/signal_width)))
+        self.dev.write("SOURce{}:FUNC:SHAPE EMEM1".format(ch))
+        self.dev.write("SOURce{}:VOLTage:LEVel:IMMediate:LOW {}".format(ch,min(data)))
+        self.dev.write("SOURce{}:VOLTage:LEVel:IMMediate:HIGH {}".format(ch,max(data)))
+        self.dev.write("SOURce{}:FREQ {}".format(ch, self.__Hz(1/signal_width)))
 
     def ext_trig(self):
         self.dev.write("TRIGger:SEQuence:SOURce EXTernal")
     
-    def burst(self, mode, cycles):
+    def burst(self, ch, mode, cycles):
         """
         Description: Allows on to use burst functionallity
 
@@ -944,11 +1069,11 @@ class tektronix_AFG3000:
         Output: 
         """
         modes = ['TRIG','GAT']
-        self.dev.write("SOURCE:BURSt:MODE {}".format(modes[mode]))
-        self.dev.write("SOURCE:BURSt:NCYC {}".format(cycles))
-        self.dev.write("SOURCE:BURSt:STAT ON")
+        self.dev.write("SOURCE{}:BURSt:MODE {}".format(ch,modes[mode]))
+        self.dev.write("SOURCE{}:BURSt:NCYC {}".format(ch,cycles))
+        self.dev.write("SOURCE{}:BURSt:STAT ON".format(ch))
     
-    def DC(self, offset):
+    def DC(self, ch, offset):
         """
         Description: Enables DC Mode (Tested 04/03/2022)
 
@@ -957,10 +1082,8 @@ class tektronix_AFG3000:
 
         Output: None : class method
         """
-        self.dev.write("SOUR:FUNC:SHAP DC")
-        self.dev.write("SOUR:VOLT:LEV:IMM:OFFS {}".format(offset))
-        
-        
+        self.dev.write("SOUR{}:FUNC:SHAP DC".format(ch))
+        self.dev.write("SOUR{}:VOLT:LEV:IMM:OFFS {}".format(ch,offset))
 
 #================================================================
 # Rigol DP8000 Series DC Power Supply
